#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import re
from pathlib import Path

try:
    import unreal
except Exception:
    print(
        "Seems like UnrealClient used outside of Unreal Editor session. Some functions may not work."
    )
    unreal = None

from typing import Optional

from .base_step_handler import BaseStepHandler


if unreal:

    @unreal.uclass()
    class RemoteRenderMoviePipelineEditorExecutor(unreal.MoviePipelinePIEExecutor):
        totalFrameRange = unreal.uproperty(int)  # Total frame range of the job's level sequence
        currentFrame = unreal.uproperty(int)  # Current frame handler that will be updating later

        def _post_init(self):
            """
            Constructor that gets called when created either via C++ or Python
            Note that this is different from the standard __init__ function of Python
            """
            self.totalFrameRange = 0
            self.currentFrame = 0

        @unreal.ufunction(override=True)
        def execute(self, queue: unreal.MoviePipelineQueue):
            """
            Execute the provided Queue.
            You are responsible for deciding how to handle each job in the queue and processing them.

            Here we define totalFrameRange as frames count from the sequence/job configuration

            :param queue: The queue that this should process all jobs for
            :return: None
            """

            # get the single job from queue
            jobs = queue.get_jobs()
            if len(jobs) == 0:
                unreal.log_error(f"Render Executor: Error: {queue} has 0 jobs")

            job = jobs[0]

            # get output settings block
            output_settings = job.get_configuration().find_or_add_setting_by_class(
                unreal.MoviePipelineOutputSetting
            )

            # if user override frame range, use overriden values
            if output_settings.use_custom_playback_range:
                self.totalFrameRange = (
                    output_settings.custom_end_frame - output_settings.custom_start_frame
                )

            # else use default frame range of the level sequence
            else:
                level_sequence = unreal.EditorAssetLibrary.load_asset(
                    unreal.SystemLibrary.conv_soft_object_reference_to_string(
                        unreal.SystemLibrary.conv_soft_obj_path_to_soft_obj_ref(job.sequence)
                    )
                )
                if level_sequence is None:
                    unreal.log_error(
                        "Render Executor: Error: Level Sequence not loaded. Check if the sequence "
                        "exists and is valid"
                    )

                self.totalFrameRange = (
                    level_sequence.get_playback_end() - level_sequence.get_playback_start()
                )

            if self.totalFrameRange == 0:
                unreal.log_error(
                    "Render Executor: Error: Cannot render the Queue with frame range of zero length"
                )

            # don't forget to call parent's execute to run the render process
            super().execute(queue)

        @unreal.ufunction(override=True)
        def on_begin_frame(self):
            """
            Called once at the beginning of each engine frame (e.g. tick, fps)
            Since the executor will work with Play in Editor widget, each rendered frame will match with widget frame tick.
            """

            super(RemoteRenderMoviePipelineEditorExecutor, self).on_begin_frame()

            # Since PIEExecutor launching Play in Editor before mrq is rendering, we should ensure, that
            # executor actually rendering the sequence.
            if self.is_rendering():
                self.currentFrame += 1
                progress = self.currentFrame / self.totalFrameRange * 100

                # Executor work with the render queue after all frames are rendered - do all
                # support stuff, handle safe quit, etc, so we should ignore progress that more than 100.
                # TODO refactor if possible, check shot/job finished callbacks
                if progress <= 100:
                    unreal.log(f"Render Executor: Progress: {progress}")


class UnrealRenderStepHandler(BaseStepHandler):
    @staticmethod
    def regex_pattern_progress() -> list[re.Pattern]:
        return [re.compile(".*Render Executor: Progress: ([0-9.]+)")]

    @staticmethod
    def regex_pattern_complete() -> list[re.Pattern]:
        return [
            re.compile(".*Render Executor: Rendering is complete"),
            re.compile(".* finished ([0-9]+) jobs in .*"),
        ]

    @staticmethod
    def regex_pattern_error() -> list[re.Pattern]:
        return [re.compile(".*Exception:.*|.*Render Executor: Error:.*|.*LogPython: Error:.*")]

    @staticmethod
    def executor_failed_callback(executor, pipeline, is_fatal, error):
        unreal.log_error(f"Render Executor: Error: {error}")

    @staticmethod
    def executor_finished_callback(movie_pipeline=None, results=None):
        unreal.log("Render Executor: Rendering is complete")

    @staticmethod
    def create_queue_from_manifest(movie_pipeline_queue_subsystem, queue_manifest_path: str):
        """
        Create the unreal.MoviePipelineQueue object from the given queue manifest path

        :param movie_pipeline_queue_subsystem: The unreal.MoviePipelineQueueSubsystem instance
        :param queue_manifest_path: Path to the manifest file
        """
        queue_manifest_path = queue_manifest_path.replace("\\", "/")
        manifest_queue = unreal.MoviePipelineLibrary.load_manifest_file_from_string(
            queue_manifest_path
        )

        pipeline_queue = movie_pipeline_queue_subsystem.get_queue()
        pipeline_queue.delete_all_jobs()
        pipeline_queue.copy_from(manifest_queue)

    @staticmethod
    def create_queue_from_job_args(
        movie_pipeline_queue_subsystem,
        level_sequence_path: str,
        level_path: str,
        job_configuration_path: str,
        job_name: Optional[str] = None,
    ):
        """
        Create the unreal.MoviePipelineQueue object from the given job arguments

        :param movie_pipeline_queue_subsystem: The unreal.MoviePipelineQueueSubsystem instance
        :param level_sequence_path: Unreal path to the level sequence file (e.g. /Game/Path/To/LevelSequence)
        :param level_path: Unreal path to the level file (e.g. /Game/Path/To/Level)
        :param job_configuration_path: Unreal path to the job configuration file (e.g. /Game/Path/To/JobConfiguration)
        :param job_name: [OPTIONAL] Name of the job to create
        """

        project_settings = unreal.get_default_object(unreal.MovieRenderPipelineProjectSettings)

        pipeline_queue = movie_pipeline_queue_subsystem.get_queue()
        pipeline_queue.delete_all_jobs()

        render_job = pipeline_queue.allocate_new_job(
            unreal.SystemLibrary.conv_soft_class_path_to_soft_class_ref(
                project_settings.default_executor_job
            )
        )

        render_job.sequence = unreal.SoftObjectPath(level_sequence_path)  # level sequence
        render_job.map = unreal.SoftObjectPath(level_path)  # level
        render_job.set_configuration(  # configuration
            unreal.EditorAssetLibrary.load_asset(job_configuration_path)
        )

        name = job_name or Path(level_sequence_path).stem
        render_job.job_name = name

    def run_script(self, args: dict) -> bool:
        """
        Create the unreal.MoviePipelineQueue object and render it with the render executor

        :param args: arguments for creating the unreal.MoviePipelineQueue object
        :return: always True, because the Unreal launch render always as async process.
            (https://docs.unrealengine.com/5.2/en-US/PythonAPI/class/MoviePipelineQueueEngineSubsystem.html#unreal.MoviePipelineQueueEngineSubsystem.render_queue_with_executor_instance)
        """
        unreal.log(f"{UnrealRenderStepHandler.run_script.__name__} executing with args: {args} ...")

        asset_registry = unreal.AssetRegistryHelpers.get_asset_registry()
        asset_registry.wait_for_completion()

        subsystem = unreal.get_editor_subsystem(unreal.MoviePipelineQueueSubsystem)

        if args.get("queue_manifest_path"):
            UnrealRenderStepHandler.create_queue_from_manifest(
                movie_pipeline_queue_subsystem=subsystem,
                queue_manifest_path=args["queue_manifest_path"],
            )
        else:
            UnrealRenderStepHandler.create_queue_from_job_args(
                movie_pipeline_queue_subsystem=subsystem,
                level_sequence_path=args.get("level_sequence_path", ""),
                level_path=args.get("level_path", ""),
                job_configuration_path=args.get("job_configuration_path", ""),
            )

        # Initialize Render executor
        executor = RemoteRenderMoviePipelineEditorExecutor()

        # Add callbacks on complete and error actions to handle it and provide output to the Deadline Adaptor
        executor.on_executor_errored_delegate.add_callable(
            UnrealRenderStepHandler.executor_failed_callback
        )
        executor.on_executor_finished_delegate.add_callable(
            UnrealRenderStepHandler.executor_finished_callback
        )

        # Render queue with the given executor
        subsystem.render_queue_with_executor_instance(executor)

        return True

    def wait_result(self, args: Optional[dict] = None) -> None:
        """
        :param args: A dictionary that contains the arguments for waiting.
        :return: None

        It is responsible for waiting result of the
        :meth:`deadline.unreal_adaptor.UnrealClient.step_handlers.unreal_render_step_handler.UnrealRenderStepHandler.run_script()`.
        """
        unreal.log("Render wait start")
        unreal.log("Render wait finish")



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
import unreal
import threading
from enum import Enum
from deadline.client.api import (
    create_job_from_job_bundle,
    get_deadline_cloud_library_telemetry_client,
)
from deadline.job_attachments.exceptions import AssetSyncCancelledError

from deadline.unreal_submitter.unreal_open_job.open_job_description import OpenJobDescription

from ._version import version


class UnrealSubmitStatus(Enum):
    """
    Enumeration of the current UnrealSubmitter status:
    - COMPLETED
    - HASHING
    - UPLOADING
    """

    COMPLETED = 1
    HASHING = 2
    UPLOADING = 3


class UnrealSubmitter:
    """
    Execute the OpenJob submission
    """

    def __init__(self, silent_mode: bool = False):
        self._silent_mode = silent_mode

        self._jobs: list[OpenJobDescription] = []
        self.submit_status: UnrealSubmitStatus = UnrealSubmitStatus.COMPLETED
        self.submit_message: str = "Start submitting..."
        self.progress_list: list[float] = []

        self.continue_submission = True  # affect all not submitted jobs
        self.submitted_job_ids: list[str] = []  # use after submit loop is ended
        self._submission_failed_message = ""  # reset after each job in the loop

        # Initialize telemetry client, opt-out is respected
        get_deadline_cloud_library_telemetry_client().update_common_details(
            {
                "deadline-cloud-for-unreal-engine-submitter-version": version,
                # TODO: record unreal-engine-version
            }
        )

    @property
    def submission_failed_message(self) -> str:
        return self._submission_failed_message

    def add_job(self, mrq_job: unreal.MoviePipelineExecutorJob):
        """
        Build and add to the submission queue :class:`deadline.unreal_submitter.unreal_open_job.open_job_description.OpenJobDescription`

        :param mrq_job: unreal.MoviePipelineExecutorJob instance
        :type mrq_job: unreal.MoviePipelineExecutorJob
        """
        self._jobs.append(OpenJobDescription(mrq_job=mrq_job))

    def _display_progress(self, check_submit_status, message):
        """
        Display the operation progress in the UI.

        :param check_submit_status: :class:`deadline.unreal_submitter.submitter.UnrealSubmitStatus` value
        :type check_submit_status: :class:`deadline.unreal_submitter.submitter.UnrealSubmitStatus`
        :param message: Message to display
        :type message: str
        """
        last_progress: float = 0
        with unreal.ScopedSlowTask(100, message) as submit_task:
            submit_task.make_dialog(True)
            while self.submit_status == check_submit_status:
                if self.submission_failed_message != "":
                    break

                if submit_task.should_cancel():
                    self.continue_submission = False
                    break

                if len(self.progress_list) > 0:
                    new_progress = self.progress_list.pop(0)
                else:
                    new_progress = last_progress
                submit_task.enter_progress_frame(new_progress - last_progress, self.submit_message)
                last_progress = new_progress

    def _start_submit(self, job_bundle_path):
        """
        Start the OpenJob submission

        :param job_bundle_path: Path of the Job bundle to submit
        :type job_bundle_path: str
        """
        try:
            job_id = create_job_from_job_bundle(
                job_bundle_dir=job_bundle_path,
                hashing_progress_callback=lambda hash_metadata: self._hash_progress(hash_metadata),
                upload_progress_callback=lambda upload_metadata: self._upload_progress(
                    upload_metadata
                ),
                create_job_result_callback=lambda: self._create_job_result(),
            )
            if job_id:
                unreal.log(f"Job creation result: {job_id}")
                self.submitted_job_ids.append(job_id)

        except AssetSyncCancelledError as e:
            unreal.log(str(e))

        except Exception as e:
            unreal.log(str(e))
            self._submission_failed_message = str(e)

    def _hash_progress(self, hash_metadata) -> bool:
        """
        Hashing progress callback for displaying hash metadata on the progress bar

        :param hash_metadata: :class:`deadline.job_attachments.progress_tracker.ProgressReportMetadata`
        :type hash_metadata: deadline.job_attachments.progress_tracker.ProgressReportMetadata
        :return: Continue submission or not
        :rtype: bool
        """
        self.submit_status = UnrealSubmitStatus.HASHING
        unreal.log(
            "Hash progress: {} {}".format(hash_metadata.progress, hash_metadata.progressMessage)
        )
        self.submit_message = hash_metadata.progressMessage
        self.progress_list.append(hash_metadata.progress)
        return self.continue_submission

    def _upload_progress(self, upload_metadata) -> bool:
        """
        Uploading progress callback for displaying upload metadata on the progress bar

        :param upload_metadata: :class:`deadline.job_attachments.progress_tracker.ProgressReportMetadata`
        :type upload_metadata: deadline.job_attachments.progress_tracker.ProgressReportMetadata
        :return: Continue submission or not
        :rtype: bool
        """

        self.submit_status = UnrealSubmitStatus.UPLOADING
        unreal.log(
            "Upload progress: {} {}".format(
                upload_metadata.progress, upload_metadata.progressMessage
            )
        )
        self.submit_message = upload_metadata.progressMessage
        self.progress_list.append(upload_metadata.progress)
        return self.continue_submission

    def _create_job_result(self) -> bool:
        """
        Creates job result callback
        :return: True
        """

        self.submit_status = UnrealSubmitStatus.COMPLETED
        unreal.log("Create job result...")
        return True

    def show_message_dialog(
        self, message: str, title="Job Submission", message_type=unreal.AppMsgType.OK
    ):
        """
        Show message dialog in the Unreal Editor UI

        :param message: Message to display
        :type message: str
        :param title: Message title
        :type title: str
        :param message_type: Message box type
        :type message_type: unreal.AppMsgType.OK
        """

        if self._silent_mode:
            return

        unreal.EditorDialog.show_message(title=title, message=message, message_type=message_type)

    def submit_jobs(self):
        """
        Submit OpenJobs to the Deadline Cloud
        """
        for job in self._jobs:
            unreal.log("Creating job from bundle...")
            self.submit_status = UnrealSubmitStatus.HASHING
            self.progress_list = []
            self.submit_message = "Start submitting..."
            self._submission_failed_message = ""

            t = threading.Thread(
                target=self._start_submit, args=(job.job_bundle_path,), daemon=True
            )
            t.start()

            self._display_progress(
                check_submit_status=UnrealSubmitStatus.HASHING, message="Hashing"
            )
            if self.continue_submission:
                self._display_progress(
                    check_submit_status=UnrealSubmitStatus.UPLOADING, message="Uploading"
                )
            t.join()

            # current job failed, notify and continue
            if self.submission_failed_message != "":
                self.show_message_dialog(
                    f"Job {job.name} unsubmitted for the reason:\n {self.submission_failed_message}"
                )

            # User cancel submission, notify and quit submission queue
            if not self.continue_submission:
                self.show_message_dialog(
                    f"Jobs submission canceled.\n"
                    f"Number of unsubmitted jobs: {len(self._jobs) - len(self.submitted_job_ids)}"
                )
                break

        # Summary notification about submission process
        self.show_message_dialog(
            f"Submitted jobs ({len(self.submitted_job_ids)}):\n" + "\n".join(self.submitted_job_ids)
        )

        del self.submitted_job_ids[:]
        del self._jobs[:]



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import unreal
from pathlib import Path


def get_project_file_path() -> str:
    """
    Returns the Unreal project OS path

    :return: the Unreal project OS path
    :rtype: str
    """

    if unreal.Paths.is_project_file_path_set():
        project_file_path = unreal.Paths.convert_relative_path_to_full(
            unreal.Paths.get_project_file_path()
        )
        return project_file_path
    else:
        raise RuntimeError("Failed to get a project name. Please set a project!")


def get_project_directory() -> str:
    """
    Returns the Unreal project directory OS path

    :return: the Unreal project directory OS path
    :rtype: str
    """

    project_file_path = get_project_file_path()
    project_directory = str(Path(project_file_path).parent).replace("\\", "/")
    return project_directory


def soft_obj_path_to_str(soft_obj_path: unreal.SoftObjectPath) -> str:
    """
    Converts the given unreal.SoftObjectPath to the Unreal path

    :param soft_obj_path: unreal.SoftObjectPath instance
    :type soft_obj_path: unreal.SoftObjectPath
    :return: the Unreal path, e.g. /Game/Path/To/Asset
    """
    obj_ref = unreal.SystemLibrary.conv_soft_obj_path_to_soft_obj_ref(soft_obj_path)
    return unreal.SystemLibrary.conv_soft_object_reference_to_string(obj_ref)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import os
import sys
from http import HTTPStatus

from typing import Optional

if "PYTHONPATH" in os.environ:
    for p in os.environ["PYTHONPATH"].split(os.pathsep):
        if p not in sys.path:
            sys.path.insert(0, p.replace("\\", "/"))

for p in sys.path:
    print(p)

from openjd.adaptor_runtime_client.win_client_interface import WinClientInterface  # noqa: E402
from deadline.unreal_adaptor.UnrealClient.step_handlers.base_step_handler import (  # noqa: E402
    BaseStepHandler,
)
from deadline.unreal_adaptor.UnrealClient.step_handlers import get_step_handler_class  # noqa: E402


class UnrealClient(WinClientInterface):
    """
    Socket DCC client implementation for UnrealEngine that send requests for actions and execute them.
    """

    def __init__(self, socket_path: str) -> None:
        super().__init__(socket_path)
        self.handler: BaseStepHandler
        self.actions.update({"set_handler": self.set_handler})

    def set_handler(self, handler_dict: dict) -> None:
        """Set the current Step Handler"""

        handler_class = get_step_handler_class(handler_dict.get("handler", "base"))
        self.handler = handler_class()
        # This is an abstract method in a base class and isn't callable but the actual handler will implement this as callable.
        # TODO: Properly type hint self.handler
        self.actions.update(self.handler.action_dict)  # type: ignore

    def close(self, args: Optional[dict] = None) -> None:
        """Close the Unreal Engine"""
        import unreal

        unreal.log("Quit the Editor: normal shutdown")
        unreal.SystemLibrary.quit_editor()

    def graceful_shutdown(self, *args, **kwargs) -> None:
        """Close the Unreal Engine if the UnrealAdaptor terminate the client with 0s grace time"""
        import unreal

        unreal.log("Quit the Editor: graceful shutdown")
        unreal.SystemLibrary.quit_editor()

    def poll(self) -> None:
        """
        This function will poll the server for the next task. If the server is in between Subtasks
        (no actions in the queue), a backoff function will be called to add a delay between the
        requests.
        """
        status, reason, action = self._request_next_action()
        if status == HTTPStatus.OK:
            if action is not None:
                print(
                    f"Performing action: {action}",
                    flush=True,
                )
                self._perform_action(action)
        else:  # Any other status or reason
            print(
                f"ERROR: An error was raised when trying to connect to the server: {status} "
                f"{reason}",
                file=sys.stderr,
                flush=True,
            )


def main():
    import unreal

    socket_path = os.environ.get("UNREAL_ADAPTOR_SOCKET_PATH", "")
    print(f"SOCKET_PATH: {socket_path}")
    if not socket_path:
        raise OSError(
            "UnrealClient cannot connect to the Adaptor because the environment variable "
            "UNREAL_ADAPTOR_SOCKET_PATH does not exist"
        )

    if not os.path.exists(socket_path):
        raise OSError(
            "UnrealClient cannot connect to the Adaptor because the socket at the path defined by "
            "the environment variable UNREAL_ADAPTOR_SOCKET_PATH does not exist. Got: "
            f"{os.environ['UNREAL_ADAPTOR_SOCKET_PATH']}"
        )

    @unreal.uclass()
    class OnTickThreadExecutorImplementation(unreal.PythonGameThreadExecutor):
        """
        Python implementation of the OnTickThreadExecutor class that runs the
        :meth:`deadline.unreal_adaptor.UnrealClient.unreal_client.UnrealClient.poll()`
        """

        client = UnrealClient(socket_path)
        time_elapsed = unreal.uproperty(float)

        def _post_init(self):
            self.time_elapsed = 0

        @unreal.ufunction(override=True)
        def execute(self, delta_time: float):
            self.time_elapsed += delta_time
            if self.time_elapsed >= 1:
                self.time_elapsed = 0
                self.client.poll()


if __name__ == "__main__":  # pragma: no cover
    main()



import os

TEMPLATES_DIRECTORY = f"{os.path.dirname(__file__)}/templates".replace("\\", "/")

DEFAULT_JOB_TEMPLATE_FILE_NAME = "default_unreal_job_template_v06.yaml"
DEFAULT_JOB_TEMPLATE_FILE_PATH = f"{TEMPLATES_DIRECTORY}/{DEFAULT_JOB_TEMPLATE_FILE_NAME}"

DEFAULT_JOB_STEP_TEMPLATE_FILE_NAME = "default_unreal_step_template_v05.yaml"
DEFAULT_JOB_STEP_TEMPLATE_FILE_PATH = f"{TEMPLATES_DIRECTORY}/{DEFAULT_JOB_STEP_TEMPLATE_FILE_NAME}"



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import sys
import threading
import unreal

import boto3
import deadline.client.config as config

from deadline.client import api
from deadline.client.api import AwsCredentialsSource, AwsAuthenticationStatus
from deadline.client.config import config_file
from deadline.job_attachments.models import FileConflictResolution


@unreal.ustruct()
class UnrealAwsEntity(unreal.StructBase):
    id = unreal.uproperty(str)
    name = unreal.uproperty(str)

    @staticmethod
    def create(entity_dict, id_field):
        """
        Create a new AwsEntity instance.

        :param entity_dict: A dictionary containing the entity data.
        :type entity_dict: dict
        :param id_field: The name of the field that represents the entity's ID.
        :type id_field: str
        :return: The newly created AwsEntity instance.
        :rtype: AwsEntity
        """
        aws_entity = UnrealAwsEntity()
        aws_entity.id = entity_dict[id_field]
        aws_entity.name = entity_dict["displayName"]
        return aws_entity


# TODO handling config parameter in api calls
def _get_current_os() -> str:
    """
    Get a string specifying what the OS is, following the format the Deadline storage profile API expects.
    """
    if sys.platform.startswith("linux"):
        return "linux"
    if sys.platform.startswith("darwin"):
        return "macos"
    if sys.platform.startswith("win"):
        return "windows"
    return "Unknown"


@unreal.uclass()
class DeadlineCloudDeveloperSettingsImplementation(unreal.DeadlineCloudDeveloperSettings):
    farms_cache_list = unreal.uproperty(unreal.Array(UnrealAwsEntity), meta=dict(Category="Cache"))
    queues_cache_list = unreal.uproperty(unreal.Array(UnrealAwsEntity), meta=dict(Category="Cache"))
    storage_profile_cache_list = unreal.uproperty(
        unreal.Array(UnrealAwsEntity), meta=dict(Category="Cache")
    )

    def _post_init(self):
        self.refresh_from_default_profile()
        self.refresh_state()

    def refresh_from_default_profile(self):
        t = threading.Thread(target=self.__refresh_deadline_settings, daemon=True)
        t.start()

    def __refresh_deadline_settings(self):
        # TODO handle case when user is not logged in
        aws_profile_name = config.get_setting("defaults.aws_profile_name")
        if aws_profile_name in ["(default)", "default", ""]:
            aws_profile_name = "(default)"

        self.work_station_configuration.global_settings.aws_profile = aws_profile_name

        self.work_station_configuration.profile.job_history_dir.path = config.get_setting(
            "settings.job_history_dir"
        ).replace("\\", "/")

        farm_id = config.get_setting("defaults.farm_id")
        farm_entity = self.find_farm_by_id(farm_id)

        if farm_entity is not None:
            self.work_station_configuration.profile.default_farm = farm_entity.name

        queue_id = config.get_setting("defaults.queue_id")
        queue_entity = self.find_queue_by_id(queue_id)
        if queue_entity is not None:
            self.work_station_configuration.farm.default_queue = queue_entity.name

        storage_profile_id = config.get_setting("settings.storage_profile_id")
        storage_profile_entity = self.find_storage_profile_by_id(storage_profile_id)
        if storage_profile_entity is not None:
            self.work_station_configuration.farm.default_storage_profile = (
                storage_profile_entity.name
            )

        self.work_station_configuration.farm.job_attachment_filesystem_options = config.get_setting(
            "defaults.job_attachments_file_system"
        )

        self.work_station_configuration.general.auto_accept_confirmation_prompts = (
            True if config.get_setting("settings.auto_accept") == "true" else False
        )

        self.work_station_configuration.general.conflict_resolution_option = config.get_setting(
            "settings.conflict_resolution"
        )

        self.work_station_configuration.general.current_logging_level = config.get_setting(
            "settings.log_level"
        )

    @unreal.ufunction(ret=unreal.Array(str))
    def get_aws_profiles(self):
        session = boto3.Session()
        aws_profile_names = list(session._session.full_config["profiles"].keys())
        for i in range(len(aws_profile_names)):
            if aws_profile_names[i] in ["default", "(defaults)", ""]:
                aws_profile_names[i] = "(default)"
        return aws_profile_names

    @unreal.ufunction(ret=unreal.Array(str))
    def get_farms(self):
        try:
            response = api.list_farms()
            self.farms_cache_list = [
                UnrealAwsEntity.create(item, "farmId") for item in response["farms"]
            ]
        # TODO: do proper exception handling
        except Exception:
            return []
        return [farm.name for farm in self.farms_cache_list]

    @unreal.ufunction(ret=unreal.Array(str))
    def get_queues(self):
        default_farm_id = config_file.get_setting("defaults.farm_id")
        if default_farm_id:
            try:
                response = api.list_queues(farmId=default_farm_id)
                self.queues_cache_list = [
                    UnrealAwsEntity.create(item, "queueId") for item in response["queues"]
                ]
            # TODO: do proper exception handling
            except Exception:
                self.queues_cache_list = []
        else:
            self.queues_cache_list = []
        return [queue.name for queue in self.queues_cache_list]

    @unreal.ufunction(ret=unreal.Array(str))
    def get_storage_profiles(self):
        default_farm_id = config_file.get_setting("defaults.farm_id")
        default_queue_id = config_file.get_setting("defaults.queue_id")
        if default_farm_id and default_queue_id:
            try:
                response = api.list_storage_profiles_for_queue(
                    farmId=default_farm_id, queueId=default_queue_id
                )
                items = response.get("storageProfiles", [])
                items.append(
                    {
                        "storageProfileId": "",
                        "displayName": "<none selected>",
                        # TODO remove _get_current_os from here
                        "osFamily": _get_current_os(),
                    }
                )
                self.storage_profile_cache_list = [
                    UnrealAwsEntity.create(item, "storageProfileId") for item in items
                ]
            # TODO: do proper exception handling
            except Exception:
                self.storage_profile_cache_list = []
            return [storage_profile.name for storage_profile in self.storage_profile_cache_list]
        else:
            self.storage_profile_cache_list = []
        return self.storage_profile_cache_list

    @unreal.ufunction(ret=unreal.Array(str))
    def get_job_attachment_modes(self):
        return ["COPIED", "VIRTUAL"]

    @unreal.ufunction(ret=unreal.Array(str))
    def get_conflict_resolution_options(self):
        return [option.name for option in FileConflictResolution]

    @unreal.ufunction(ret=unreal.Array(str))
    def get_logging_levels(self):
        return ["ERROR", "WARNING", "INFO", "DEBUG"]

    @unreal.ufunction(override=True)
    def refresh_state(self):
        t = threading.Thread(target=self.__refresh_deadline_status, daemon=True)
        t.start()

    def __refresh_deadline_status(self):
        config_parser = config_file.read_config()
        self.work_station_configuration.state.creds_type = api.get_credentials_source(
            config=config_parser
        ).name

        creds_status = api.check_authentication_status(config=config_parser)
        self.work_station_configuration.state.creds_status = creds_status.name

        if creds_status == AwsAuthenticationStatus.AUTHENTICATED:
            self.work_station_configuration.state.api_availability = (
                "AUTHORIZED"
                if api.check_deadline_api_available(config=config_parser)
                else "NOT AUTHORIZED"
            )
        else:
            self.work_station_configuration.state.api_availability = "NOT AUTHORIZED"

    @unreal.ufunction(override=True)
    def on_settings_modified(self, property_name):
        """
        TODO refresh config
        1. When we change an "aws profile" we need to pull Job history dir and default farm
        2. When we change "default farm" we need to pull default queue,
           default storage profile, and job attachment fs options
        """
        unreal.log(f"Changed property: {property_name}")

        # This means we need to change default profile
        # If the default profile is changed then we update it in the config first
        # after that we need to save this setting first and read settings for all other values
        if property_name == "AWS_Profile":
            config.set_setting(
                "defaults.aws_profile_name",
                self.work_station_configuration.global_settings.aws_profile,
            )
            self.refresh_from_default_profile()
            return

        if property_name == "DefaultFarm":
            farm = self.find_farm_by_name(self.work_station_configuration.profile.default_farm)
            if farm is not None:
                config.set_setting("defaults.farm_id", farm.id)
            self.refresh_from_default_profile()
            return

        self.save_to_file()

    def find_farm_by_id(self, farm_id):
        _ = self.get_farms()
        farm = next((farm for farm in self.farms_cache_list if farm.id == farm_id), None)
        return farm

    def find_queue_by_id(self, queue_id):
        _ = self.get_queues()
        queue = next((queue for queue in self.queues_cache_list if queue.id == queue_id), None)
        return queue

    def find_storage_profile_by_id(self, storage_profile_id):
        _ = self.get_storage_profiles()
        storage_profile = next(
            (
                storage_profile
                for storage_profile in self.storage_profile_cache_list
                if storage_profile.id == storage_profile_id
            ),
            None,
        )
        return storage_profile

    def find_farm_by_name(self, farm_name):
        # response = api.list_farms()
        farm = next((farm for farm in self.farms_cache_list if farm.name == farm_name), None)
        return farm

    def find_queue_by_name(self, queue_name):
        # default_farm_id = config_file.get_setting("defaults.farm_id")
        # response = api.list_queues(farmId=default_farm_id)
        queue = next((queue for queue in self.queues_cache_list if queue.name == queue_name), None)
        return queue

    def find_storage_by_name(self, storage_profile_name):
        storage_profile = next(
            (item for item in self.storage_profile_cache_list if item.name == storage_profile_name),
            None,
        )
        return storage_profile

    def save_to_file(self):
        config_parser = config_file.read_config()
        config.set_setting(
            "defaults.aws_profile_name",
            self.work_station_configuration.global_settings.aws_profile,
            config=config_parser,
        )

        config.set_setting(
            "settings.job_history_dir",
            self.work_station_configuration.profile.job_history_dir.path,
            config=config_parser,
        )

        farm = self.find_farm_by_name(self.work_station_configuration.profile.default_farm)
        if farm is not None:
            unreal.log(f"Update default farm: {farm.id} -- {farm.name}")
            config.set_setting("defaults.farm_id", farm.id, config=config_parser)

        queue = self.find_queue_by_name(self.work_station_configuration.farm.default_queue)
        if queue is not None:
            unreal.log(f"Update default queue: {queue.id} -- {queue.name}")
            config.set_setting("defaults.queue_id", queue.id, config=config_parser)

        storage_profile = self.find_storage_by_name(
            self.work_station_configuration.farm.default_storage_profile
        )
        if storage_profile is not None:
            unreal.log(
                f"Update default storage profile: {storage_profile.id} "
                f"-- {storage_profile.name}"
            )
            config.set_setting(
                "settings.storage_profile_id", storage_profile.id, config=config_parser
            )

        # farm.job_attachment_filesystem_options (defaults.job_attachments_file_system)
        config.set_setting(
            "defaults.job_attachments_file_system",
            self.work_station_configuration.farm.job_attachment_filesystem_options,
            config=config_parser,
        )

        if self.work_station_configuration.general.auto_accept_confirmation_prompts:
            config.set_setting("settings.auto_accept", "true", config=config_parser)
        else:
            config.set_setting("settings.auto_accept", "false", config=config_parser)

        # general.conflict_resolution_option (settings.conflict_resolution)
        config.set_setting(
            "settings.conflict_resolution",
            self.work_station_configuration.general.conflict_resolution_option,
            config=config_parser,
        )

        # general.current_logging_level
        config.set_setting(
            "settings.log_level",
            self.work_station_configuration.general.current_logging_level,
            config=config_parser,
        )

        config_file.write_config(config_parser)

    @unreal.ufunction(override=True)
    def login(self):
        unreal.log("login")

        def on_pending_authorization(**kwargs):
            if kwargs["credential_type"] == AwsCredentialsSource.DEADLINE_CLOUD_MONITOR_LOGIN:
                unreal.EditorDialog.show_message(
                    "Deadline Cloud",
                    "Opening Deadline Cloud Monitor. Please login before returning here.",
                    unreal.AppMsgType.OK,
                    unreal.AppReturnType.OK,
                )

        def on_cancellation_check():
            return False

        success_message = api.login(
            on_pending_authorization,
            on_cancellation_check,
            config=None,
        )
        if success_message:
            unreal.EditorDialog.show_message(
                "Deadline Cloud", success_message, unreal.AppMsgType.OK, unreal.AppReturnType.OK
            )
            self.refresh_from_default_profile()
            self.refresh_state()

    @unreal.ufunction(override=True)
    def logout(self):
        unreal.log("Deadline Cloud logout")
        api.logout()
        self.refresh_from_default_profile()
        self.refresh_state()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
import unreal
from deadline.unreal_submitter.submitter import UnrealSubmitter


@unreal.uclass()
class MoviePipelineDeadlineCloudRemoteExecutor(unreal.MoviePipelineExecutorBase):
    # The queue we are working on, null if no queue has been provided.
    pipeline_queue = unreal.uproperty(unreal.MoviePipelineQueue)
    job_ids = unreal.uproperty(unreal.Array(str))

    @unreal.ufunction(override=True)
    def execute(self, pipeline_queue):
        unreal.log(f"Asked to execute Queue: {pipeline_queue}")
        unreal.log(f"Queue has {len(pipeline_queue.get_jobs())} jobs")

        if not pipeline_queue or (not pipeline_queue.get_jobs()):
            self.on_executor_finished_impl()
            return

        if not self.check_dirty_packages():
            return

        if not self.check_maps(pipeline_queue):
            return

        self.pipeline_queue = pipeline_queue

        # TODO how do they use it?
        # deadline_settings = unreal.get_default_object(
        #     unreal.DeadlineCloudRenderStepSetting
        # )

        # TODO Custom commandline arguments

        unreal_submitter = UnrealSubmitter()

        for job in self.pipeline_queue.get_jobs():
            unreal.log(f"Submitting Job `{job.job_name}` to Deadline Cloud...")
            unreal_submitter.add_job(job)

        unreal_submitter.submit_jobs()

    @unreal.ufunction(override=True)
    def is_rendering(self):
        # Because we forward unfinished jobs onto another service when the
        # button is pressed, they can always submit what is in the queue and
        # there's no need to block the queue.
        # A MoviePipelineExecutor implementation must override this. If you
        # override a ufunction from a base class you don't specify the return
        # type or parameter types.
        return False

    def check_dirty_packages(self) -> bool:
        dirty_packages = []
        dirty_packages.extend(unreal.EditorLoadingAndSavingUtils.get_dirty_content_packages())
        dirty_packages.extend(unreal.EditorLoadingAndSavingUtils.get_dirty_map_packages())

        if dirty_packages:
            if not unreal.EditorLoadingAndSavingUtils.save_dirty_packages_with_dialog(True, True):
                message = (
                    "One or more jobs in the queue have an unsaved map/content. "
                    "{packages} "
                    "Please save and check-in all work before submission.".format(
                        packages="\n".join(dirty_packages)
                    )
                )

                unreal.log_error(message)
                unreal.EditorDialog.show_message(
                    "Unsaved Maps/Content", message, unreal.AppMsgType.OK
                )
                self.on_executor_finished_impl()
                return False
        return True

    def check_maps(self, pipeline_queue) -> bool:
        has_valid_map = unreal.MoviePipelineEditorLibrary.is_map_valid_for_remote_render(
            pipeline_queue.get_jobs()
        )
        if not has_valid_map:
            message = (
                "One or more jobs in the queue have an unsaved map as "
                "their target map. "
                "These unsaved maps cannot be loaded by an external process, "
                "and the render has been aborted."
            )
            unreal.log_error(message)
            unreal.EditorDialog.show_message("Unsaved Maps", message, unreal.AppMsgType.OK)
            self.on_executor_finished_impl()
            return False

        return True



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import unreal
from typing import Callable, Optional

from .common import os_path_from_unreal_path
from .dependency_search_options import DependencySearchOptions

asset_registry = unreal.AssetRegistryHelpers.get_asset_registry()


class DependencyCollector:
    """
    A helper class to collect all dependencies of the given unreal asset (Level, LevelSequence, etc.).
    Execute recursive collecting on the newly found dependency  until not found any other.

    For example, we want to collect dependencies of LevelSequence:
    1. LevelSequence depends on Level and Cube
    2. Level depends on StatueSet
    3. StatueSet depends on HorseAsset, RockAsset
    Output list will be: [Level, Cube, StatueSet, HorseAsset, RockAsset]
    """

    def __init__(self):
        self._collected_dependencies = list()

    def collect(
        self,
        asset_path: str,
        dependency_options=DependencySearchOptions(),
        filter_method: Optional[Callable] = None,
        on_found_dependency_callback: Optional[Callable] = None,
    ):
        """
        Collect all dependencies recursively of the given unreal asset.

        :param asset_path: Unreal path of the asset to find dependencies, e.g. /Game/Sequences/MyLevelSequence
        :type asset_path: str
        :param dependency_options: Dataclass containing options for search dependency
        :type dependency_options: DependencySearchOptions
        :param filter_method: Method used to filter the found dependencies, for example, dependencies only in Game(Content) folder
        :type filter_method: typing.Callable, optional
        :param on_found_dependency_callback: Method used to invoke some operations on found dependencies list, for example sync them from VCS
        :type on_found_dependency_callback: typing.Callable, optional

        :return: List of the collected dependencies
        :rtype: list
        """
        self._collected_dependencies.clear()

        udependency_options = unreal.AssetRegistryDependencyOptions(**dependency_options.as_dict())

        source_control_available = unreal.SourceControl.is_available()
        unreal.log(
            "DependencyCollector: Source control is available: {}".format(source_control_available)
        )

        if source_control_available:
            if not unreal.EditorAssetLibrary.does_asset_exist(asset_path):
                os_asset_path = os_path_from_unreal_path(asset_path) + ".*"
                unreal.SourceControl.sync_file(os_asset_path)
                unreal.AssetRegistryHelpers().get_asset_registry().scan_modified_asset_files(
                    [asset_path]
                )
                unreal.AssetRegistryHelpers().get_asset_registry().scan_paths_synchronous(
                    [asset_path], True, True
                )

        dependencies = self._get_dependencies(
            asset_path=asset_path,
            udependency_options=udependency_options,
            filter_method=filter_method,
            on_found_dependency_callback=on_found_dependency_callback,
        )

        return dependencies

    def _get_dependencies(
        self,
        asset_path: str,
        udependency_options: unreal.AssetRegistryDependencyOptions,
        filter_method: Optional[Callable] = None,
        on_found_dependency_callback: Optional[Callable] = None,
    ):
        """
        Inner method that gets called recursively and execute the main collecting process

        :param asset_path: Unreal path of the asset to find dependencies, e.g. /Game/Sequences/MyLevelSequence
        :type asset_path: str
        :param udependency_options: Asset Registry Dependency Options (https://docs.unrealengine.com/5.2/en-US/PythonAPI/class/AssetRegistryDependencyOptions.html)
        :type udependency_options: unreal.AssetRegistryDependencyOptions
        :param filter_method: Method used to filter the found dependencies, for example, dependencies only in Game(Content) folder
        :type filter_method: typing.Callable, optional
        :param on_found_dependency_callback: Method used to invoke some operations on found dependencies list, for example sync them from VCS
        :type on_found_dependency_callback: typing.Callable, optional

        :return: List of dependencies
        :rtype: list
        """

        dependencies_raw = asset_registry.get_dependencies(
            package_name=asset_path, dependency_options=udependency_options
        )

        dependencies = list()
        if dependencies_raw:
            for dependency_raw in dependencies_raw:
                does_confirm_filter = filter_method(dependency_raw) if filter_method else True
                is_not_collected = dependency_raw not in self._collected_dependencies

                if does_confirm_filter and is_not_collected:
                    dependencies.append(dependency_raw)

        if dependencies:
            self._collected_dependencies.extend(dependencies)

        if on_found_dependency_callback:
            unreal.log(
                f"Execute callable {on_found_dependency_callback.__name__} on the dependencies"
            )
            on_found_dependency_callback(dependencies)

        unreal.AssetRegistryHelpers().get_asset_registry().scan_modified_asset_files(dependencies)
        unreal.AssetRegistryHelpers().get_asset_registry().scan_paths_synchronous(
            dependencies, True, True
        )

        for dependency in dependencies:
            self._get_dependencies(
                dependency, udependency_options, filter_method, on_found_dependency_callback
            )

        return [str(d) for d in self._collected_dependencies]



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import re
from typing import Optional
from abc import abstractmethod, ABC


class BaseStepHandler(ABC):
    def __init__(self):
        self.action_dict = dict(run_script=self.run_script, wait_result=self.wait_result)

    @abstractmethod
    def run_script(self, args: dict) -> bool:
        """
        :param args: A dictionary that contains the arguments for running the script.
        :return: boolean indicating the script run successfully or not.

        This method is an abstract method that needs to be implemented by the extending class.
        It is responsible for executing a script using the provided arguments.
        """
        raise NotImplementedError("Abstract method, need to be implemented")

    @abstractmethod
    def wait_result(self, args: Optional[dict] = None) -> None:
        """
        :param args: A dictionary that contains the arguments for waiting.
        :return: None

        This method is an abstract method that needs to be implemented by the extending class.
        It is responsible for waiting result of the
        :meth:`deadline.unreal_adaptor.UnrealClient.step_handlers.base_step_handler.BaseStepHandler.run_script()`.
        """
        raise NotImplementedError("Abstract method, need to be implemented")

    @staticmethod
    @abstractmethod
    def regex_pattern_progress() -> list[re.Pattern]:
        """Returns a list of regex Patterns that match the progress messages"""
        raise NotImplementedError("Abstract method, need to be implemented")

    @staticmethod
    @abstractmethod
    def regex_pattern_complete() -> list[re.Pattern]:
        """Returns a list of regex Patterns that match the complete messages"""
        raise NotImplementedError("Abstract method, need to be implemented")

    @staticmethod
    @abstractmethod
    def regex_pattern_error() -> list[re.Pattern]:
        """Returns a list of regex Patterns that match the errors messages"""
        raise NotImplementedError("Abstract method, need to be implemented")



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import os
import unreal


content_dir = unreal.Paths.project_content_dir()
content_dir = unreal.Paths.convert_relative_path_to_full(content_dir)
project_dir = unreal.Paths.project_dir()
project_dir = unreal.Paths.convert_relative_path_to_full(project_dir)


class DependencyFilters:
    """
    Common class for filtering the found dependencies
    """

    @staticmethod
    def dependency_in_game_folder(dependency_path):
        """
        Check if a given dependency is in the "Game" folder ("Content" in the OS file system)

        :param dependency_path: Unreal Path of the dependency asset, e.g. /Game/Assets/MyAsset
        :type dependency_path: str

        :return: True if the dependency is in the Game folder, False otherwise
        :rtype: bool
        """

        return "/Game/" in str(dependency_path)


def os_path_from_unreal_path(unreal_path, with_ext: bool = False):
    """
    Convert Unreal path to OS path, e.g. /Game/Assets/MyAsset to C:/UE_project/Content/Assets/MyAsset.uasset.

    if parameter with_ext is set to True, tries to get type of the asset by unreal.AssetData and set appropriate extension:

    - type World - .umap
    - other types - .uasset

    If for some reason it can't find asset data (e.g. temporary level's actors don't have asset data), it will set ".uasset"

    :param unreal_path: Unreal Path of the asset, e.g. /Game/Assets/MyAsset
    :param with_ext: if True, build the path with extension (.uasset or .umap), set asterisk "*" otherwise.
    :return: the OS path of the asset
    :rtype: str
    """
    os_path = str(unreal_path).replace("/Game/", content_dir)

    if with_ext:
        asset_data = unreal.EditorAssetLibrary.find_asset_data(unreal_path)
        asset_class_name = (
            asset_data.asset_class_path.asset_name
            if hasattr(asset_data, "asset_class_path")
            else asset_data.asset_class
        )  # support older version of UE python API

        if (
            not asset_class_name.is_none()
        ):  # AssetData not found - asset not in the project / on disk
            os_path += ".umap" if asset_class_name == "World" else ".uasset"
        else:
            os_path += ".uasset"
    else:
        os_path += ".*"

    return os_path


def os_abs_from_relative(os_path):
    if os.path.isabs(os_path):
        return str(os_path)
    return project_dir + os_path


def sync_assets_with_ue_source_control(asset_paths: list[str], sync_description="Sync Assets"):
    """
    Sync the given assets with Unreal Source Control plugin, which handle all connection parameters.

    :param asset_paths: Asset paths to sync
    :type asset_paths: list[str]
    :param sync_description: Sync description for the UI progress bar
    :type sync_description: str
    :return: Sync success result (True/False)
    :rtype: bool
    """
    synced = True

    if not unreal.SourceControl.is_available():
        unreal.log("SourceControl is not available")
        return

    unreal.log("Sync assets: {}".format(asset_paths))
    if "IS_RENDER_MODE" not in os.environ:
        with unreal.ScopedSlowTask(len(asset_paths), sync_description) as slow_task:
            slow_task.make_dialog(True)
            for i in range(len(asset_paths)):
                unreal.log("Sync asset: {}".format(asset_paths[i]))
                synced = unreal.SourceControl.sync_files([asset_paths[i]])
                slow_task.enter_progress_frame(1, asset_paths[i])
    else:
        synced = unreal.SourceControl.sync_files(asset_paths)

    unreal.log(f"Assets synced: {synced}")
    return synced



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from dataclasses import dataclass, asdict


@dataclass
class DependencySearchOptions:
    """
    A dataclass for storing dependencies search options

    include_hard_package_references (True by default)- Dependencies which are required for correct usage of the source asset, and must be loaded at the same time
    include_soft_package_references (True by default)- Dependencies which don't need to be loaded for the object to be used (i.e. soft object paths)
    include_hard_management_references (True by default)- Reference that says one object directly manages another object, set when Primary Assets manage things explicitly
    include_soft_management_references (True by default)- Indirect management references, these are set through recursion for Primary Assets that manage packages or other primary assets
    include_searchable_names (False by default)- References to specific SearchableNames inside a package
    """

    include_hard_package_references: bool = True
    include_soft_package_references: bool = True
    include_hard_management_references: bool = True
    include_soft_management_references: bool = True
    include_searchable_names: bool = False

    def as_dict(self):
        """
        Represent fields of dataclass as dictionary

        :return: A dictionary from its attributes
        :rtype: dict
        """
        return asdict(self)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import re
import sys
import inspect
import importlib
import traceback
from pathlib import Path
from typing import Optional
from types import ModuleType

from .base_step_handler import BaseStepHandler


class UnrealCustomStepHandler(BaseStepHandler):
    @staticmethod
    def regex_pattern_progress() -> list[re.Pattern]:
        return [re.compile(".*Custom Step Executor: Progress: ([0-9.]+)")]

    @staticmethod
    def regex_pattern_complete() -> list[re.Pattern]:
        return [re.compile(".*Custom Step Executor: Complete")]

    @staticmethod
    def regex_pattern_error() -> list[re.Pattern]:
        return [re.compile(".*Exception:.*|.*Custom Step Executor: Error:.*")]

    @staticmethod
    def validate_script(script_path: str) -> ModuleType:
        """
        This method is responsible for validating the script

        :param script_path: Path of the script to validate
        :return: If script is valid, returns its as module, None otherwise
        """

        _script_path = Path(script_path)

        if not _script_path.exists() or not _script_path.is_file():
            raise FileNotFoundError(f"Script {script_path} does not exist or it is not a file")

        sys.path.append(str(_script_path.parent))
        script_module = importlib.import_module(_script_path.stem)

        has_main_method = False

        for name, obj in inspect.getmembers(script_module, predicate=inspect.isfunction):
            if name == "main":
                has_main_method = True
                break

        if not has_main_method:
            raise Exception("Invalid script. Please check the script have the 'main' method.")

        return script_module

    def run_script(self, args: dict) -> bool:
        """
        Executing a script using the provided arguments.

        :param args: A dictionary that contains the arguments for running the script.
        :return: boolean indicating the script run successfully or not.
        """

        import unreal

        try:
            script_module = UnrealCustomStepHandler.validate_script(script_path=args["script_path"])
            script_args = args.get("script_args", {})
            result = script_module.main(**script_args)
            unreal.log(f"Custom Step Executor: Complete: {result}")
            return True
        except Exception as e:
            unreal.log(
                f"Custom Step Executor: Error: "
                f'Error occured while executing the given script {args.get("script_path")}: {str(e)}\n'
            )
            unreal.log(traceback.format_exc())
            return False

    def wait_result(self, args: Optional[dict] = None) -> None:
        """
        :param args: A dictionary that contains the arguments for waiting.
        :return: None

        It is responsible for waiting result of the
        :meth:`deadline.unreal_adaptor.UnrealClient.step_handlers.unreal_custom_step_handler.UnrealCustomStepHandler.run_script()`.
        """
        import unreal

        unreal.log("Render wait start")
        unreal.log("Render wait finish")



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

from __future__ import annotations

import os
import re
import sys
import time
import logging
import threading
from typing import Callable

from deadline.client.api import get_deadline_cloud_library_telemetry_client, TelemetryClient
from openjd.adaptor_runtime._version import version as openjd_adaptor_version
from openjd.adaptor_runtime_client import Action
from openjd.adaptor_runtime.process import LoggingSubprocess
from openjd.adaptor_runtime.adaptors import Adaptor, SemanticVersion
from openjd.adaptor_runtime.app_handlers import RegexCallback, RegexHandler
from openjd.adaptor_runtime.application_ipc import ActionsQueue, AdaptorServer
from openjd.adaptor_runtime.adaptors.configuration import AdaptorConfiguration

from .._version import version as adaptor_version
from .common import DataValidation, add_module_to_pythonpath

logger = logging.getLogger(__name__)


class UnrealNotRunningError(Exception):
    """Error that is raised when attempting to use Unreal while it is not running"""

    pass


class UnrealSubprocessWithLogs(LoggingSubprocess): ...


class UnrealAdaptor(Adaptor[AdaptorConfiguration]):
    """
    Adaptor that creates a session in Unreal to Render interactively.
    """

    _SERVER_START_TIMEOUT_SECONDS = 30
    _SERVER_END_TIMEOUT_SECONDS = 30
    _UNREAL_START_TIMEOUT_SECONDS = 86400
    _UNREAL_END_TIMEOUT_SECONDS = 30

    _server: AdaptorServer | None = None

    _server_thread: threading.Thread | None = None

    _unreal_client: UnrealSubprocessWithLogs | None = None

    _action_queue = ActionsQueue()

    _is_rendering: bool = False

    _exc_info: Exception | None = None

    _performing_cleanup = False

    _telemetry_client: TelemetryClient | None = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.data_validation = DataValidation()

    @property
    def integration_data_interface_version(self) -> SemanticVersion:
        return SemanticVersion(major=0, minor=1)

    @staticmethod
    def get_timer(timeout: int | float) -> Callable[[], bool]:
        """Given a timeout length, returns a lambda which returns True until the timeout occurs"""
        timeout_time = time.time() + timeout
        return lambda: time.time() < timeout_time

    @property
    def _has_exception(self) -> bool:
        """Property which checks the private _exc_info property for an exception

        :raises self._exc_info: An exception if there is one

        :return: False there is no exception waiting to be raised
        :rtype: bool
        """
        if self._exc_info and not self._performing_cleanup:
            raise self._exc_info
        return False

    @property
    def _unreal_is_running(self) -> bool:
        """Property which indicates that the unreal client is running

        :return: True if the unreal client is running, false otherwise
        :rtype: bool
        """
        return self._unreal_client is not None and self._unreal_client.is_running

    @property
    def _unreal_is_rendering(self) -> bool:
        """Property which indicates if unreal is rendering

        :return: True if unreal is rendering, false otherwise
        :rtype: bool
        """
        return self._unreal_is_running and self._is_rendering

    @_unreal_is_rendering.setter
    def _unreal_is_rendering(self, value: bool) -> None:
        """Property setter which updates the private _is_rendering boolean.

        :param bool value: A boolean indicated if unreal is rendering.
        """
        self._is_rendering = value

    @property
    def unreal_client_path(self) -> str:
        """
        Obtains the unreal_client.py path by searching directories in sys.path

        :raises FileNotFoundError: If the unreal_client.py file could not be found.

        :return: The path to the unreal_client.py file.
        :rtype: str
        """
        for p in sys.path:
            path = os.path.join(p, "deadline", "unreal_adaptor", "UnrealClient", "unreal_client.py")
            if os.path.isfile(path):
                return path
        raise FileNotFoundError(
            "Could not find unreal_client.py. Check that the UnrealClient package is in one of the "
            f"following directories: {sys.path[1:]}"
        )

    def _wait_for_adaptor_server_socket(self) -> str:
        """
        Performs a busy wait for the socket path that the adaptor server is running on, then
        returns it.

        :raises RuntimeError: If the server does not finish initializing

        :return: The socket path the adaptor server is running on.
        :rtype: str
        """
        is_not_timed_out = self.get_timer(self._SERVER_START_TIMEOUT_SECONDS)

        while (self._server is None or self._server.server_path is None) and is_not_timed_out():
            time.sleep(0.01)

        if self._server is not None and self._server.server_path is not None:
            return self._server.server_path

        raise RuntimeError(
            "Could not find a socket path because the server did not finish initializing"
        )

    def _wait_for_unreal_started(self):
        """
        Performs a busy wait for the starting of the Unreal Engine with the UnrealClient script

        :raises RuntimeError: Raised when the UnrealClient encountered an error during initialization
        :raises TimeoutError: Raised when the UnrealClient doesn't complete the initial actions before timeout reached
        """
        is_not_timed_out = self.get_timer(self._UNREAL_START_TIMEOUT_SECONDS)
        while (
            self._unreal_is_running
            and not self._has_exception
            and len(self._action_queue)
            > 0  # for now the initializing actions in the action queue, defined by
            # _populate_action_queue() method.
            # So we wait for them to be done or for time is out.
            and is_not_timed_out()
        ):
            time.sleep(0.1)

        self._get_deadline_telemetry_client().record_event(
            event_type="com.amazon.rum.deadline.adaptor.runtime.start", event_details={}
        )

        if len(self._action_queue) > 0:  # if for some reason, all the actions are not complete
            if is_not_timed_out():  # and timeout is not reached
                raise RuntimeError(  # <- we catch some exception - self._has_exception is True
                    "Unreal encountered an error and was not able to complete initialization actions."
                )
            else:
                raise TimeoutError(  # All the actions are note complete and timeout reached
                    "Unreal did not complete initialization actions in "
                    f"{self._UNREAL_START_TIMEOUT_SECONDS} seconds and failed to start."
                )

    def _start_unreal_server(self) -> None:
        """
        Starts a server with the given ActionsQueue, attaches the server to the adaptor and serves
        forever in a blocking call.
        """
        self._server = AdaptorServer(self._action_queue, self)
        self._server.serve_forever()

    def _start_unreal_server_thread(self) -> None:
        """
        Starts the unreal adaptor server in a thread.
        Sets the environment variable "UNREAL_ADAPTOR_SOCKET_PATH" to the socket the server is running
        on after the server has finished starting.
        """

        self._server_thread = threading.Thread(
            target=self._start_unreal_server, name="UnrealAdaptorServerThread"
        )
        self._server_thread.start()
        os.environ["UNREAL_ADAPTOR_SOCKET_PATH"] = self._wait_for_adaptor_server_socket()

    def _get_regex_callbacks(self) -> list[RegexCallback]:
        """
        Returns a list of RegexCallbacks built from UnrealClient handler regex patterns
        :return: List of Regex Callbacks to add
        :rtype: list[RegexCallback]
        """

        callbacks = []

        from deadline.unreal_adaptor.UnrealClient.step_handlers import get_step_handler_class

        for handler_name in ["render", "custom"]:
            handler_class = get_step_handler_class(handler_name)

            logger.info(f"Gettings regex pattertns from step handler: {handler_class}...")

            callbacks.append(
                RegexCallback(handler_class.regex_pattern_progress(), self._handle_progress)
            )

            callbacks.append(
                RegexCallback(handler_class.regex_pattern_complete(), self._handle_complete)
            )

            callbacks.append(RegexCallback(handler_class.regex_pattern_error(), self._handle_error))

        return callbacks

    def _handle_complete(self, match: re.Match) -> None:
        """
        Callback for stdout that indicate completeness of a render. Updates progress to 100

        :param match: re.Match object from the regex pattern that was matched the message
        :type match: re.Match
        """
        self._unreal_is_rendering = False
        self.update_status(progress=100)

    def _handle_progress(self, match: re.Match) -> None:
        """
        Callback for stdout that indicate progress of a render.

        :param match: re.Match object from the regex pattern that was matched the message
        :type match: re.Match
        """
        progress = int(float(match.groups()[0]))
        self.update_status(progress=progress)

    def _handle_error(self, match: re.Match) -> None:
        """
        Callback for stdout that indicates an error or warning.

        :param match: re.Match object from the regex pattern that was matched the message
        :type match: re.Match

        :raises RuntimeError: Always raises a runtime error to halt the adaptor.
        """
        self._exc_info = RuntimeError(f"Unreal Encountered an Error: {match.group(0)}")

    def _start_unreal_client(self) -> None:
        """
        Starts the Unreal client by launching UnrealEditor-Cmd with the unreal_client.py file.

        UnrealEditor-Cmd must be on the system PATH, for example due to a Rez environment being active.

        :raises FileNotFoundError: If the unreal_client.py file could not be found.
        """

        unreal_exe = "UnrealEditor-Cmd"
        unreal_project_path = self.init_data.get("project_path", "")
        client_path = self.unreal_client_path.replace("\\", "/")
        log_args = [
            "-log",
            "-unattended",
            "-stdout",
            "-NoLoadingScreen",
            "-NoScreenMessages",
            "-RenderOffscreen",
            "-allowstdoutlogverbosity",
        ]

        args = [unreal_exe, unreal_project_path]
        args.extend(log_args)
        args.append(f"-execcmds=r.HLOD 0,py {client_path}")

        regexhandler = RegexHandler(self._get_regex_callbacks())
        self._unreal_client = UnrealSubprocessWithLogs(
            args=args,
            stdout_handler=regexhandler,
            stderr_handler=regexhandler,
        )

    def _populate_action_queue(self) -> None:
        """
        Populates the adaptor server's action queue with actions from the init_data that the Unreal
        Client will request and perform.
        """
        pass
        # Set up all pathmapping rules
        # self._action_queue.enqueue_action(
        #     Action(
        #         "path_mapping",
        #         {
        #             "path_mapping_rules": {
        #                 rule.source_path: rule.destination_path for rule in self.path_mapping_rules
        #             }
        #         },
        #     )
        # )

    def _get_deadline_telemetry_client(self):
        """
        Wrapper around the Deadline Client Library telemetry client, in order to set package-specific information
        """
        if not self._telemetry_client:
            self._telemetry_client = get_deadline_cloud_library_telemetry_client()
            self._telemetry_client.update_common_details(
                {
                    "deadline-cloud-for-unreal-engine-adaptor-version": adaptor_version,
                    "open-jd-adaptor-runtime-version": openjd_adaptor_version,
                    # TODO: add unreal-engine-version
                }
            )
        return self._telemetry_client

    def on_start(self) -> None:
        """
        For job stickiness. Will start everything required for the Task.

        :raises:
            jsonschema.ValidationError: When init_data fails validation against the adaptor schema.
            jsonschema.SchemaError: When the adaptor schema itself is nonvalid.
            RuntimeError: If Unreal did not complete initialization actions due to an exception
            TimeoutError: If Unreal did not complete initialization actions due to timing out.
            FileNotFoundError: If the unreal_client.py file could not be found.
        """

        self.data_validation.validate_init_data(self.init_data)

        # Notify worker agent about starting Unreal
        self.update_status(progress=0, status_message="Initializing Unreal Engine")

        # Starts the unreal adaptor server
        self._start_unreal_server_thread()

        self._populate_action_queue()

        # Add the openjd and adaptor namespace directory to PYTHONPATH, so that adaptor_runtime_client
        # will be available directly to the adaptor client.

        import openjd.adaptor_runtime_client

        add_module_to_pythonpath(
            os.path.dirname(os.path.dirname(openjd.adaptor_runtime_client.__file__))
        )
        add_module_to_pythonpath(
            os.path.dirname(
                os.path.dirname(os.path.dirname(openjd.adaptor_runtime_client.__file__))
            )
        )

        import deadline.unreal_adaptor

        add_module_to_pythonpath(os.path.dirname(os.path.dirname(deadline.unreal_adaptor.__file__)))

        self._start_unreal_client()

        self._wait_for_unreal_started()

    def on_run(self, run_data: dict) -> None:
        """
        This starts a render in Unreal for the given frame and performs a busy wait until the render completes.

        :param run_data: Dictionary containing Run Data
        :type run_data: dict
        """
        if not self._unreal_is_running:
            raise UnrealNotRunningError("Cannot render because Unreal is not running.")

        self.data_validation.validate_run_data(run_data)

        # Set up the step handler
        self._action_queue.enqueue_action(
            Action("set_handler", {"handler": run_data.get("handler", "base")})
        )

        self._unreal_is_rendering = True
        self._action_queue.enqueue_action(Action("run_script", run_data))

        while self._unreal_is_rendering and not self._has_exception:
            time.sleep(1)
            # Wait for:
            #   1. set_handler to be executed
            #   2. run_script to be executed (launch render process in separate thread by UE API)
            # by UnrealClient and don't spam before
            if len(self._action_queue) == 0:
                logger.info("Enqueue wait result")
                self._action_queue.enqueue_action(Action("wait_result", {}))

        if (
            not self._unreal_is_running and self._unreal_client
        ):  # Unreal Client will always exist here.
            #  This is always an error case because the Unreal Client should still be running and
            #  waiting for the next command. If the thread finished, then we cannot continue
            exit_code = self._unreal_client.returncode
            self._get_deadline_telemetry_client().record_error(
                {"exit_code": exit_code, "exception_scope": "on_run"}, str(RuntimeError)
            )
            raise RuntimeError(
                "Unreal exited early and did not render successfully, please check render logs. "
                f"Exit code {exit_code}"
            )

    def on_stop(self) -> None:
        """
        Execute stop action
        """
        pass

    def on_cleanup(self):
        """
        Cleans up the adaptor by closing the unreal client and adaptor server.
        """

        self._performing_cleanup = True

        # Send "close" action to the UnrealClient
        self._action_queue.enqueue_action(Action("close"), front=True)

        # Wait for UnrealClient to be terminated
        is_not_timed_out = self.get_timer(self._UNREAL_END_TIMEOUT_SECONDS)
        while self._unreal_is_running and is_not_timed_out():
            time.sleep(0.1)

        if self._unreal_is_running and self._unreal_client:
            logger.error(
                "Unreal did not complete cleanup actions and failed to gracefully shutdown. "
                "Terminating."
            )
            self._unreal_client.terminate(0)

        # Terminate AdaptorServer instance
        if self._server:
            self._server.shutdown()

        # Terminate AdaptorServer thread launched in background
        if self._server_thread and self._server_thread.is_alive():
            self._server_thread.join(timeout=self._SERVER_END_TIMEOUT_SECONDS)
            if self._server_thread.is_alive():
                logger.error("Failed to shutdown the Unreal Adaptor server.")

        self._performing_cleanup = False

    def on_cancel(self):
        """
        Cancels the current render if Unreal is rendering.
        """
        logger.info("CANCEL REQUESTED")
        if not self._unreal_client or not self._unreal_is_running:
            logger.info("Nothing to cancel because Unreal is not running")
            return

        # Terminate immediately since the Unreal client does not have a graceful shutdown
        self._unreal_client.terminate(grace_time_s=0)



import os

import unreal

from deadline.unreal_submitter.common import soft_obj_path_to_str
from deadline.unreal_submitter.unreal_dependency_collector.collector import DependencyCollector
from deadline.unreal_submitter.unreal_dependency_collector.common import (
    DependencyFilters,
    os_path_from_unreal_path,
)

# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.


@unreal.uclass()
class DeadlineCloudJobBundleLibraryImplementation(unreal.DeadlineCloudJobBundleLibrary):
    @unreal.ufunction(override=True)
    def get_job_dependencies(self, mrq_job):
        level_sequence_path = soft_obj_path_to_str(mrq_job.sequence)
        level_path = soft_obj_path_to_str(mrq_job.map)

        level_sequence_path, _ = os.path.splitext(level_sequence_path)
        level_path, _ = os.path.splitext(level_path)

        dependency_collector = DependencyCollector()
        unreal.log("Level sequence: " + level_sequence_path)
        unreal.log("Level: " + level_path)

        unreal_dependencies = dependency_collector.collect(
            asset_path=level_sequence_path,
            filter_method=DependencyFilters.dependency_in_game_folder,
        )

        unreal_dependencies += dependency_collector.collect(
            asset_path=level_path, filter_method=DependencyFilters.dependency_in_game_folder
        )

        unreal_dependencies += [level_sequence_path, level_path]

        unreal.log(
            f"Converted level path: {os_path_from_unreal_path(level_sequence_path, with_ext=True)}"
        )

        unreal_dependencies = list(set(unreal_dependencies))

        return [os_path_from_unreal_path(d, with_ext=True) for d in unreal_dependencies]

    @unreal.ufunction(override=True)
    def get_cpu_architectures(self):
        return ["x86_64", "arm64"]

    @unreal.ufunction(override=True)
    def get_operating_systems(self):
        return ["linux", "macos", "windows"]

    @unreal.ufunction(override=True)
    def get_job_initial_state_options(self):
        return ["READY", "SUSPENDED"]



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
from typing import Type

from .base_step_handler import BaseStepHandler
from .unreal_render_step_handler import UnrealRenderStepHandler
from .unreal_custom_step_handler import UnrealCustomStepHandler

__all__ = ["BaseStepHandler", "get_step_handler_class"]


def get_step_handler_class(handler: str = "base") -> Type[BaseStepHandler]:
    """
    Returns the step handler instance for the given handler name.

    Args:
        handler (str, optional): The handler name to get the handler instance of.
            Defaults to "BaseStepHandler".

    Returns the BaseStepHandler instance for the given handler name.
    """

    handlers_map = dict(
        base=BaseStepHandler, render=UnrealRenderStepHandler, custom=UnrealCustomStepHandler
    )

    print(
        f'Trying to get step handler class, defined as "{handler}" in the handlers map: {handlers_map}'
    )
    handler_class = handlers_map.get(handler, BaseStepHandler)
    print(f"Got step handler class: {handler_class}")

    return handler_class



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import logging
import sys

from openjd.adaptor_runtime import EntryPoint

from .adaptor import UnrealAdaptor

__all__ = ["main"]
_logger = logging.getLogger(__name__)


def main(reentry_exe=None) -> int:
    _logger.info("About to start the UnrealAdaptor")

    package_name = vars(sys.modules[__name__])["__package__"]
    if not package_name:
        raise RuntimeError(f"Must be run as a module. Do not run {__file__} directly")

    try:
        EntryPoint(UnrealAdaptor).start(reentry_exe=reentry_exe)
    except Exception as e:
        _logger.error(f"Entrypoint failed: {e}")
        return 1

    _logger.info("Done UnrealAdaptor main")
    return 0


if __name__ == "__main__":
    sys.exit(main())



import os
import yaml
import unreal
from copy import deepcopy
from typing import Dict, Any, List

from deadline.client.job_bundle import deadline_yaml_dump, create_job_history_bundle_dir
from deadline.client.job_bundle.submission import AssetReferences
from deadline.unreal_submitter.settings import DEFAULT_JOB_TEMPLATE_FILE_PATH
from deadline.unreal_submitter.common import (
    get_project_directory,
    get_project_file_path,
    soft_obj_path_to_str,
)
from deadline.unreal_submitter.unreal_dependency_collector.common import (
    DependencyFilters,
    os_path_from_unreal_path,
)
from deadline.unreal_submitter.unreal_dependency_collector.common import os_abs_from_relative
from deadline.unreal_submitter.unreal_dependency_collector.collector import DependencyCollector

from deadline.unreal_submitter.unreal_open_job.job_step import JobStep, JobStepFactory


class JobSharedSettings:
    """
    OpenJob shared settings representation.
    Contains SharedSettings model as dictionary built from template and allows to fill its values
    """

    def __init__(self, job_shared_settings):
        self.source_shared_settings = job_shared_settings
        self.parameter_values: List[Dict[Any, Any]] = [
            {
                "name": "deadline:targetTaskRunStatus",
                "type": "STRING",
                "userInterface": {
                    "control": "DROPDOWN_LIST",
                    "label": "Initial State",
                },
                "allowedValues": ["READY", "SUSPENDED"],
                "value": self.get_initial_state(),
            },
            {
                "name": "deadline:maxFailedTasksCount",
                "description": "Maximum number of Tasks that can fail before the Job will be marked as failed.",
                "type": "INT",
                "userInterface": {
                    "control": "SPIN_BOX",
                    "label": "Maximum Failed Tasks Count",
                },
                "minValue": 0,
                "value": self.get_max_failed_tasks_count(),
            },
            {
                "name": "deadline:maxRetriesPerTask",
                "description": "Maximum number of times that a Task will retry before it's marked as failed.",
                "type": "INT",
                "userInterface": {
                    "control": "SPIN_BOX",
                    "label": "Maximum Retries Per Task",
                },
                "minValue": 0,
                "value": self.get_max_retries_per_task(),
            },
            {"name": "deadline:priority", "type": "INT", "value": self.get_priority()},
        ]

    def to_dict(self) -> list[dict]:
        """
        Returns the OpenJob SharedSettings object as list of dictionaries

        :return: OpenJob SharedSettings as list of dictionaries
        :rtype: dict
        """
        return self.parameter_values

    def get_initial_state(self) -> str:
        """
        Returns the OpenJob Initial State value

        :return: OpenJob Initial State
        :rtype: str
        """
        return self.source_shared_settings.initial_state

    def get_max_failed_tasks_count(self) -> int:
        """
        Returns the OpenJob Max Failed Task Count value

        :return: OpenJob Max Failed Task Count
        :rtype: int
        """
        return self.source_shared_settings.maximum_failed_tasks_count

    def get_max_retries_per_task(self) -> int:
        """
        Returns the OpenJob Max Retries Per Task value

        :return: OpenJob Max Retries Per Task
        :rtype: int
        """
        return self.source_shared_settings.maximum_retries_per_task

    def get_priority(self) -> int:
        """
        Return the OpenJob Priority value

        :return: OpenJob Priority
        :rtype: int
        """
        # TODO Add priority to the settings
        return 1


class OpenJobDescription:
    """
    Represents a OpenJob description object
    """

    def __init__(self, mrq_job: unreal.MoviePipelineExecutorJob):
        """
        Build OpenJob with the given MovieP ipeline Executor Job and Queue Manifest path

        :param mrq_job: unreal.MoviePipelineExecutorJob object with the Job context
        :type mrq_job: unreal.MoviePipelineExecutorJob
        :param manifest_path: Path to the QueueManifest file with the Job parameters
        :type manifest_path: str
        """
        with open(DEFAULT_JOB_TEMPLATE_FILE_PATH) as f:
            self.default_job_template = yaml.safe_load(f)

        self._dependency_collector = DependencyCollector()

        self._open_job: Dict
        self._manifest_path: str

        self._steps: list[JobStep] = []
        self._parameter_values_dict: Dict[Any, Any] = {}
        self._asset_references = AssetReferences()
        self._job_bundle_path: str

        self._create_open_job_from_mrq_job(mrq_job)

    @property
    def name(self):
        """
        Returns the OpenJob name
        """
        return self._open_job.get("name")

    @property
    def job_bundle_path(self):
        """
        Returns the OpenJob job bundle path
        """
        return self._job_bundle_path

    def _create_open_job_from_mrq_job(self, mrq_job: unreal.MoviePipelineExecutorJob) -> None:
        """
        Creates an OpenJob representation from the unreal.MoviePipelineExecutorJob.

        Set the name, build steps, fill parameter values, attachments and build final job bundle.

        :param mrq_job: unreal.MoviePipelineExecutorJob instance
        :type mrq_job: unreal.MoviePipelineExecutorJob
        """

        self._open_job = deepcopy(self.default_job_template)
        shared_settings = mrq_job.preset_overrides.job_shared_settings

        self._open_job["name"] = (
            mrq_job.job_name
            if shared_settings.name == "" or shared_settings.name == "Untitled"
            else shared_settings.name
        )

        self._open_job["description"] = shared_settings.description

        self._save_manifest_file(mrq_job)

        self._build_steps(mrq_job)
        self._open_job["steps"] = [step.as_dict() for step in self._steps]

        self._build_parameter_values_dict(mrq_job)
        self._build_asset_references(mrq_job)

        self._build_job_bundle()

    def _collect_mrq_job_dependencies(self, mrq_job) -> list[str]:
        """
        Collects the dependencies of the Level and LevelSequence that used in MRQ Job.

        Use :class:`deadline.unreal_submitter.unreal_dependency_collector.collector.DependencyCollector` for collecting

        :param mrq_job: unreal.MoviePipelineExecutorJob instance
        :type mrq_job: unreal.MoviePipelineExecutorJob

        :return: List of the dependencies
        :rtype: list[str]
        """
        level_sequence_path = soft_obj_path_to_str(mrq_job.sequence)
        level_sequence_path = os.path.splitext(level_sequence_path)[0]

        level_path = soft_obj_path_to_str(mrq_job.map)
        level_path = os.path.splitext(level_path)[0]

        level_sequence_dependencies = self._dependency_collector.collect(
            level_sequence_path, filter_method=DependencyFilters.dependency_in_game_folder
        )

        level_dependencies = self._dependency_collector.collect(
            level_path, filter_method=DependencyFilters.dependency_in_game_folder
        )

        return level_sequence_dependencies + level_dependencies + [level_sequence_path, level_path]

    def _build_parameter_values_dict(self, mrq_job: unreal.MoviePipelineExecutorJob) -> dict:
        """
        Build parameter values of the OpenJob with the given MRQ Job.
        Extend the built parameter values with OpenJob SharedSettings instance
        (:class:`deadline.unreal_submitter.unreal_open_job.open_job_description.JobSharedSettings`)

        :param mrq_job: unreal.MoviePipelineExecutorJob instance
        :type mrq_job: unreal.MoviePipelineExecutorJob

        :return: Parameter values dictionary
        :rtype: dict
        """

        project_file_path = get_project_file_path()
        project_directory = get_project_directory()

        # Output path
        output_setting = mrq_job.get_configuration().find_setting_by_class(
            unreal.MoviePipelineOutputSetting
        )
        output_path = output_setting.output_directory.path
        # TODO handling unreal substitution templates
        output_path = output_path.replace("{project_dir}", project_directory)

        parameter_values = [
            {
                "name": "LevelPath",
                "value": soft_obj_path_to_str(mrq_job.map),
            },
            {
                "name": "LevelSequencePath",
                "value": soft_obj_path_to_str(mrq_job.sequence),
            },
            {
                "name": "ProjectFilePath",
                "value": project_file_path,
            },
            {
                "name": "ProjectDirectory",
                "value": project_directory,
            },
            {"name": "OutputPath", "value": output_path},
        ]

        shared_parameter_values = JobSharedSettings(
            mrq_job.preset_overrides.job_shared_settings
        ).to_dict()
        parameter_values += shared_parameter_values

        self._parameter_values_dict = dict(parameterValues=parameter_values)

        return self._parameter_values_dict

    def _build_asset_references(self, mrq_job) -> AssetReferences:
        """
        Build asset references of the OpenJob with the given MRQ Job.

        Return :class:`deadline.client.job_bundle.submission.AssetReferences` instance

        :param mrq_job: unreal.MoviePipelineExecutorJob instance
        :type mrq_job: unreal.MoviePipelineExecutorJob

        :return: AssetReferences dataclass instance
        :rtype: :class:`deadline.client.job_bundle.submission.AssetReferences`
        """

        # add dependencies to attachments
        os_dependencies = []
        job_dependencies = self._collect_mrq_job_dependencies(mrq_job)
        for dependency in job_dependencies:
            os_dependency = os_path_from_unreal_path(dependency, with_ext=True)
            if os.path.exists(os_dependency):
                os_dependencies.append(os_dependency)

        self._asset_references.input_filenames.update(os_dependencies)

        step_input_files = []
        for step in self._steps:
            step_input_files.extend(step.get_step_input_files())
        self._asset_references.input_filenames.update(step_input_files)

        project_directory = get_project_directory()

        # add manifest to attachments
        self._asset_references.input_filenames.add(self._manifest_path)

        # add other input files to attachments
        job_input_files = [
            os_abs_from_relative(input_file.file_path)
            for input_file in mrq_job.preset_overrides.job_attachments.input_files.files.paths
        ]
        for input_file in job_input_files:
            if os.path.exists(input_file):
                self._asset_references.input_filenames.add(input_file)

        # required input directories
        for sub_dir in ["Config", "Binaries"]:
            input_directory = os_abs_from_relative(sub_dir)
            if os.path.exists(input_directory):
                self._asset_references.input_directories.add(input_directory)

        # input directories
        job_input_directories = [
            os_abs_from_relative(input_directory.path)
            for input_directory in mrq_job.preset_overrides.job_attachments.input_directories.directories.paths
        ]
        for input_dir in job_input_directories:
            if os.path.exists(input_dir):
                self._asset_references.input_directories.add(input_dir)

        # output directories
        job_output_directories = [
            os_abs_from_relative(output_directory.path)
            for output_directory in mrq_job.preset_overrides.job_attachments.output_directories.directories.paths
        ]
        for output_dir in job_output_directories:
            self._asset_references.input_directories.add(output_dir)

        # TODO handling unreal substitution templates (output directory)
        # TODO Think about it. Maybe we need to move files from Unreal output to Deadline cloud output on worker side
        # or we should override output directory with the output defined in DataAsset
        output_setting = mrq_job.get_configuration().find_setting_by_class(
            unreal.MoviePipelineOutputSetting
        )
        output_path = output_setting.output_directory.path
        output_path = output_path.replace("{project_dir}", project_directory).rstrip("/")
        self._asset_references.output_directories.update([output_path])

        return self._asset_references

    def _build_steps(self, mrq_job) -> list[JobStep]:
        """
        Build OpenJob steps with the given MRQ Job.

        :param mrq_job: unreal.MoviePipelineExecutorJob instance
        :type mrq_job: unreal.MoviePipelineExecutorJob

        :return: List of the :class:`deadline.unreal_submitter.unreal_open_job.job_step.JobStep`
        :rtype: :class:`deadline.unreal_submitter.unreal_open_job.job_step.JobStep`
        """

        preset_overrides: unreal.DeadlineCloudJobPresetStruct = mrq_job.preset_overrides
        unreal.log(f"Preset overrides: {preset_overrides}")

        try:
            self._steps = JobStepFactory.create_steps(
                job_settings=mrq_job.get_configuration().get_all_settings(),
                host_requirements=preset_overrides.host_requirements,
                queue_manifest_path=self._manifest_path,
            )
            return self._steps

        except Exception as e:
            unreal.EditorDialog.show_message(
                "Custom step validate failed", str(e), unreal.AppMsgType.OK
            )
            raise e

    def _build_job_bundle(self) -> str:
        """
        Convert OpenJob to the bundle and write it on the disk.

        :return: OpenJob bundle path
        :rtype: str
        """

        job_bundle_path = create_job_history_bundle_dir("Unreal", self._open_job["name"])
        unreal.log(f"Job bundle path: {job_bundle_path}")

        with open(job_bundle_path + "/template.yaml", "w", encoding="utf8") as f:
            deadline_yaml_dump(self._open_job, f, indent=1)

        with open(job_bundle_path + "/parameter_values.yaml", "w", encoding="utf8") as f:
            deadline_yaml_dump(self._parameter_values_dict, f, indent=1)

        with open(job_bundle_path + "/asset_references.yaml", "w", encoding="utf8") as f:
            deadline_yaml_dump(self._asset_references.to_dict(), f, indent=1)

        self._job_bundle_path = job_bundle_path

        return self._job_bundle_path

    def _save_manifest_file(self, mrq_job):
        new_queue = unreal.MoviePipelineQueue()
        new_job = new_queue.duplicate_job(mrq_job)

        # In duplicated job remove empty auto-detected files since we don't want them to be saved in manifest
        # List of the files is moved to OpenJob attachments
        new_job.preset_overrides.job_attachments.input_files.auto_detected = (
            unreal.DeadlineCloudFileAttachmentsArray()
        )

        (
            duplicated_queue,
            manifest_path,
        ) = unreal.MoviePipelineEditorLibrary.save_queue_to_manifest_file(new_queue)
        manifest_path = unreal.Paths.convert_relative_path_to_full(manifest_path)
        self._manifest_path = manifest_path



import os
import yaml
import unreal
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Optional

from deadline.unreal_submitter.settings import DEFAULT_JOB_STEP_TEMPLATE_FILE_PATH
from deadline.unreal_submitter.unreal_dependency_collector.common import os_abs_from_relative


class HostRequirements:
    """OpenJob host requirements representation"""

    def __init__(self, host_requirements):
        self.source_host_requirements = host_requirements
        self.requirements: dict = {}

        os_requirements = self._get_os_requirements()
        if os_requirements:
            # OS requirements are currently all amount type capabilities
            self.requirements["attributes"] = os_requirements

        hardware_requirements = self._get_hardware_requirements()
        if hardware_requirements:
            # hardware requirements are currently all amount
            self.requirements["amounts"] = hardware_requirements

    def _get_os_requirements(self) -> list[dict]:
        """
        Get requirements for OS family and CPU architecture

        :return: list of the OS requirements
        :rtype: list[dict]
        """
        requirements: list[dict] = []
        if self.source_host_requirements.operating_system:
            requirements.append(
                {
                    "name": "attr.worker.os.family",
                    "anyOf": [self.source_host_requirements.operating_system],
                }
            )
        if self.source_host_requirements.cpu_architecture:
            requirements.append(
                {
                    "name": "attr.worker.cpu.arch",
                    "anyOf": [self.source_host_requirements.cpu_architecture],
                }
            )
        return requirements

    def _get_hardware_requirements(self) -> list[dict[str, Any]]:
        """
        Get requirements for cpu, gpu and memory limits

        :return: list of the OS requirements
        :rtype: list[dict]
        """
        cpus = self._get_amount_requirement(
            self.source_host_requirements.cp_us, "amount.worker.vcpu"
        )
        memory = self._get_amount_requirement(
            self.source_host_requirements.memory, "amount.worker.memory", 1024
        )
        # TODO gpu amount
        # gpus = self._get_amount_requirement(self.source_host_requirements.gpus, "amount.worker.gpu")
        gpu_memory = self._get_amount_requirement(
            self.source_host_requirements.gp_us, "amount.worker.gpu.memory", 1024
        )
        scratch_space = self._get_amount_requirement(
            self.source_host_requirements.scratch_space, "amount.worker.disk.scratch"
        )
        requirements: list[dict[str, Any]] = [
            item for item in [cpus, memory, gpu_memory, scratch_space] if item is not None
        ]
        return requirements

    @staticmethod
    def _get_amount_requirement(source_interval, name: str, scaling_factor: int = 1) -> dict:
        """
        Helper method to get the amount of Host Requirement setting interval

        :param source_interval: Interval unreal setting
        :param name: AWS HostRequirements setting name
        :param scaling_factor: Multiplier number by which to scale the source_interval values

        :return: Amount requirement as dictionary
        :rtype: dict
        """
        requirement = {}
        if source_interval.min > 0 or source_interval.max > 0:
            requirement = {"name": name}
            if source_interval.min > 0:
                requirement["min"] = source_interval.min * scaling_factor
            if source_interval.max > 0:
                requirement["max"] = source_interval.max * scaling_factor
        return requirement

    def as_dict(self) -> dict:
        """
        Returns the HostRequirements as dictionary

        :return: Host Requirements as dictionary
        :rtype: dict
        """
        return self.requirements


class JobStep:
    """
    Represents a OpenJob Step
    """

    def __init__(self, step_template, step_settings, host_requirements, queue_manifest_path):
        """
        Build JobStep, set its name and fill dependencies list

        :param step_template: Step default template to use for this step
        :type step_template: dict
        :param step_settings: Deadline Cloud Step Setting object
        """
        self._job_step = deepcopy(step_template)

        self._set_name(step_settings)
        self._fill_step_dependency_list(step_settings)
        self._fill_host_requirements(host_requirements)

    def _set_name(self, step_settings):
        """
        Set the name of this Step

        :param step_settings: Deadline Cloud Step Setting object with the name
        :raises Exception: When job name is empty or None
        """
        new_name = step_settings.name
        if not new_name:
            raise Exception("Job step name could not be empty")

        self._job_step["name"] = new_name

    def _set_step_path_parameter(self, parameter_name: str, path_value: str):
        """
        Fill the path parameter by name of this Step with the given value

        :param parameter_name: Name of the parameter
        :type parameter_name: str
        :param path_value: Value of the parameter
        :type path_value: str
        """
        parameter_space = self._job_step["parameterSpace"]
        parameter_definition = next(
            (
                parameter_definition
                for parameter_definition in parameter_space["taskParameterDefinitions"]
                if parameter_definition["name"] == parameter_name
            ),
            None,
        )
        if parameter_definition is not None:
            parameter_definition["range"] = [path_value]

    def _fill_step_dependency_list(self, step_settings):
        """
        Fill the dependsOn list of this Step with the given step settings' "depends_on" attribute

        :param step_settings: Deadline Cloud Step Setting object with the depends_on attribute
        """
        dependencies_list = [
            {"dependsOn": dependency} for dependency in step_settings.depends_on if dependency != ""
        ]
        if len(dependencies_list) > 0:
            self._job_step["dependencies"] = dependencies_list

    def _fill_host_requirements(self, host_requirements):
        if host_requirements.run_on_all_worker_nodes:
            return
        self._job_step["hostRequirements"] = HostRequirements(host_requirements).as_dict()

    def get_step_input_files(self) -> list[str]:
        return []

    def as_dict(self):
        """Returns a dictionary representation of this Step"""
        return self._job_step


class CustomScriptJobStep(JobStep):
    """
    Represents a OpenJob Step for Custom Script executing
    """

    def __init__(self, step_template, step_settings, host_requirements, queue_manifest_path):
        """
        Build JobStep, set its name, fill dependencies list and set script path parameter
        """
        super().__init__(step_template, step_settings, host_requirements, queue_manifest_path)

        self._set_script_path_parameter(os_abs_from_relative(step_settings.script.file_path))

    def _set_script_path_parameter(self, script_path):
        """
        Fill the necessary parameter "ScriptPath" with the given script path.

        Use :meth:`deadline.unreal_submitter.unreal_open_job.job_step.JobStep._set_step_path_parameter()`

        :param script_path: Path to the script
        :type script_path: str
        """
        if not os.path.exists(script_path):
            raise Exception(f"Script path does not exist on the disk: {script_path}")

        self._set_step_path_parameter(parameter_name="ScriptPath", path_value=script_path)

    def get_step_input_files(self) -> list[str]:
        """
        Return the script paths from ScriptPath range attribute

        :return: List of script paths
        :rtype: list[str]
        """
        script_attachments = []

        parameter_space = self._job_step["parameterSpace"]
        parameter_definition = next(
            (
                parameter_definition
                for parameter_definition in parameter_space["taskParameterDefinitions"]
                if parameter_definition["name"] == "ScriptPath"
            ),
            None,
        )
        if parameter_definition is not None:
            script_attachments = [attachment for attachment in parameter_definition["range"]]

        return script_attachments


class RenderJobStep(JobStep):
    """
    Represents a OpenJob Step for Render executing
    """

    def __init__(self, step_template, step_settings, host_requirements, queue_manifest_path):
        """
        Build JobStep, set its name, fill dependencies list and set queue manifest path parameter
        """
        super().__init__(step_template, step_settings, host_requirements, queue_manifest_path)

        self._set_queue_manifest_path_parameter(queue_manifest_path)

    def _set_name(self, step_settings):
        """
        Override the behavior of the JobStep._set_name() method and setup name as "Render"
        """
        self._job_step["name"] = "Render"

    def _set_queue_manifest_path_parameter(self, queue_manifest_path):
        """
        Fill the necessary parameter "QueueManifestPath" with the given script path.

        Use :meth:`deadline.unreal_submitter.unreal_open_job.job_step.JobStep._set_step_path_parameter`

        :param script_path: Path to the script
        :type script_path: str
        """
        self._set_step_path_parameter(
            parameter_name="QueueManifestPath", path_value=queue_manifest_path
        )


@dataclass
class JobStepDescriptor:
    """
    Common Job Step descriptor that contains information
    about the step type, its representation class and appropriate Unreal setting class
    """

    #: Step default type, for example "Render" or "CustomScript"
    step_type: str
    #: Representation class of the Step template, inherited from :class:``
    step_class: type[JobStep]
    setting_class: unreal.Class


class JobStepFactory:
    """Build JobStep with the given context"""

    with open(DEFAULT_JOB_STEP_TEMPLATE_FILE_PATH) as f:
        #: Job Step default template that contains unfilled description of the Job's steps
        DEFAULT_JOB_STEP_TEMPLATE = yaml.safe_load(f)

    #: Common Step mapping list of the :class:`deadline.unreal_submitter.unreal_open_job.job_step.JobStepDescriptor`
    #: instances for the Render and CustomScript steps
    JOB_STEP_MAPPING = [
        JobStepDescriptor(
            "Render", RenderJobStep, unreal.DeadlineCloudRenderStepSetting().static_class()
        ),
        JobStepDescriptor(
            "CustomScript",
            CustomScriptJobStep,
            unreal.DeadlineCloudCustomScriptStepSetting().static_class(),
        ),
    ]

    @staticmethod
    def get_step_descriptor_by_setting_class(
        setting_class: unreal.Class,
    ) -> Optional[JobStepDescriptor]:
        """
        Returns the :class:`deadline.unreal_submitter.unreal_open_job.job_step.JobStepDescriptor` instance that kept in
        :class:`deadline.unreal_submitter.unreal_open_job.job_step.JobStepFactory.JOB_STEP_MAPPING`
        by the given unreal setting class (Render or CustomScript)

        :param setting_class: Unreal Settings instance class
        :type setting_class: unreal.Class

        :return: :class:`deadline.unreal_submitter.unreal_open_job.job_step.JobStepDescriptor` instance for the Render/CustomScript
        :rtype: :class:`deadline.unreal_submitter.unreal_open_job.job_step.JobStepDescriptor`
        """
        return next(
            (
                step_descriptor
                for step_descriptor in JobStepFactory.JOB_STEP_MAPPING
                if step_descriptor.setting_class == setting_class
            ),
            None,
        )

    @staticmethod
    def get_step_template(step_descriptor: JobStepDescriptor):
        """
        Read the JobStep template file and get the appropriate JobStep

        :param step_descriptor: :class:`deadline.unreal_submitter.unreal_open_job.job_step.JobStepDescriptor` instance
        :type step_descriptor: :class:`deadline.unreal_submitter.unreal_open_job.job_step.JobStepDescriptor`

        :return: JobStep template
        :rtype: dict
        """

        step_template = next(
            (
                step_template
                for step_template in JobStepFactory.DEFAULT_JOB_STEP_TEMPLATE["steps"]
                if step_template["name"] == step_descriptor.step_type
            ),
            None,
        )
        return step_template

    @classmethod
    def create_steps(
        cls,
        job_settings: list[unreal.MoviePipelineSetting],
        queue_manifest_path: str,
        host_requirements,
    ) -> list[JobStep]:
        """
        Create the Job Steps list using the provided job settings and other parameters

        :param job_settings: list of unreal.MoviePipelineSetting settings
        :type job_settings: unreal.MoviePipelineSetting
        :param queue_manifest_path: OS path for the queue manifest file
        :type queue_manifest_path: str
        :param host_requirements: AWS Host requirements settings

        :return: list of the :class:`deadline.unreal_submitter.unreal_open_job.job_step.JobStep` instances
        :rtype: :class:`deadline.unreal_submitter.unreal_open_job.job_step.JobStep`
        """

        steps = []

        for setting in job_settings:
            step_descriptor = JobStepFactory.get_step_descriptor_by_setting_class(
                setting.get_class()
            )
            if not step_descriptor:
                continue

            if hasattr(setting, "deadline_cloud_steps"):
                for script_step_setting in setting.deadline_cloud_steps:
                    steps.append(
                        step_descriptor.step_class(
                            step_template=JobStepFactory.get_step_template(step_descriptor),
                            step_settings=script_step_setting,
                            host_requirements=host_requirements,
                            queue_manifest_path=queue_manifest_path,
                        )
                    )

            else:
                steps.append(
                    step_descriptor.step_class(
                        step_template=JobStepFactory.get_step_template(step_descriptor),
                        step_settings=setting,
                        host_requirements=host_requirements,
                        queue_manifest_path=queue_manifest_path,
                    )
                )

        return steps



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import os

from openjd.adaptor_runtime.adaptors import AdaptorDataValidators


def add_module_to_pythonpath(module_path: str):
    """
    Extend or create env variable PYTHONPATH and add there the given path
    """

    # can be passed the __init__.py file or the parent directory
    module_directory = os.path.dirname(module_path) if os.path.isfile(module_path) else module_path

    if "PYTHONPATH" in os.environ:
        os.environ["PYTHONPATH"] = f'{os.environ["PYTHONPATH"]}{os.pathsep}{module_directory}'
    else:
        os.environ["PYTHONPATH"] = module_directory


class DataValidation:
    """
    Common class for validating init and run data
    """

    def __init__(self):
        cur_dir = os.path.dirname(__file__)
        schema_dir = os.path.join(cur_dir, "schemas")
        self.validators = AdaptorDataValidators.for_adaptor(schema_dir)

    def validate_init_data(self, init_data: dict):
        """
        Validate the given init data

        :param init_data: Initial adaptor data
        :type init_data: dict

        :raises jsonschema.exceptions.ValidationError: if the instance is invalid
        :raises: jsonschema.exceptions.SchemaError: if the schema itself is invalid
        """
        self.validators.init_data.validate(init_data)

    def validate_run_data(self, run_data: dict):
        """
        Validate the given run data

        :raises jsonschema.exceptions.ValidationError: if the instance is invalid
        :raises jsonschema.exceptions.SchemaError: if the schema itself is invalid
        """

        self.validators.run_data.validate(run_data)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import os
import sys
import unreal

remote_execution = os.getenv("REMOTE_EXECUTION", "False")
if remote_execution != "True":
    unreal.log("INIT DEADLINE CLOUD")

    libraries_path = f"{os.path.dirname(__file__)}/libraries".replace("\\", "/")
    if not os.getenv("DEADLINE_CLOUD") and os.path.exists(libraries_path):
        os.environ["DEADLINE_CLOUD"] = libraries_path

    unreal.log(f'DEADLINE CLOUD PATH: {os.getenv("DEADLINE_CLOUD")}')
    if os.getenv("DEADLINE_CLOUD") and os.environ["DEADLINE_CLOUD"] not in sys.path:
        sys.path.append(os.environ["DEADLINE_CLOUD"])

    # These unused imports are REQUIRED!!!
    # Unreal Engine loads any init_unreal.py it finds in its search paths.
    # These imports finish the setup for the plugin.
    from settings import DeadlineCloudDeveloperSettingsImplementation  # noqa: F401
    from job_library import DeadlineCloudJobBundleLibraryImplementation  # noqa: F401
    import remote_executor  # noqa: F401
